@@ -911,6 +911,69 @@ static int string_cmp_ic(OnigEncoding enc,
 
 #define DATA_ENSURE_CHECK(n)   (s + (n) <= end)
 
+#ifdef RUBY_PLATFORM
+
+typedef struct {
+  int        state;
+  regex_t*   reg;
+  MatchArg*  msa;
+  StackType* stk_base;
+} TrapEnsureArg;
+
+static VALUE
+trap_ensure(VALUE arg)
+{
+  TrapEnsureArg* ta = (TrapEnsureArg* )arg;
+
+  if (ta->state == 0) { /* trap_exec() is not normal return */
+    ta->reg->state--;
+    if (! IS_NULL(ta->msa->stack_p) && ta->stk_base != ta->msa->stack_p)
+      xfree(ta->stk_base);
+
+    MATCH_ARG_FREE(*(ta->msa));
+  }
+
+  return Qnil;
+}
+
+static VALUE
+trap_exec(VALUE arg)
+{
+  TrapEnsureArg* ta;
+
+  rb_trap_exec();
+
+  ta = (TrapEnsureArg* )arg;
+  ta->state = 1; /* normal return */
+  return Qnil;
+}
+
+extern void
+onig_exec_trap(regex_t* reg, MatchArg* msa, StackType* stk_base)
+{
+  VALUE arg;
+  TrapEnsureArg ta;
+
+  ta.state    = 0;
+  ta.reg      = reg;
+  ta.msa      = msa;
+  ta.stk_base = stk_base;
+  arg = (VALUE )(&ta);
+  rb_ensure(trap_exec, arg, trap_ensure, arg);
+}
+
+#define CHECK_INTERRUPT_IN_MATCH_AT do {\
+  if (rb_trap_pending) {\
+    if (! rb_prohibit_interrupt) {\
+      onig_exec_trap(reg, msa, stk_base);\
+    }\
+  }\
+} while (0)
+#else
+#define CHECK_INTERRUPT_IN_MATCH_AT
+#endif /* RUBY_PLATFORM */
+
+
 #ifdef ONIG_DEBUG_STATISTICS
 
 #define USE_TIMEOFDAY
@@ -1053,6 +1116,7 @@ match_at(regex_t* reg, UChar* str, UChar* end, UChar* sstart,
   char *alloca_base;
   StackType *stk_alloc, *stk_base, *stk, *stk_end;
   StackType *stkp; /* used as any purpose. */
+  StackIndex si;
   StackIndex *repeat_stk;
   StackIndex *mem_start_stk, *mem_end_stk;
   n = reg->num_repeat + reg->num_mem * 2;
@@ -2026,6 +2090,8 @@ match_at(regex_t* reg, UChar* str, UChar* end, UChar* sstart,
 	    break;
 	  case OP_REPEAT_INC:
 	  case OP_REPEAT_INC_NG:
+	  case OP_REPEAT_INC_SG:
+	  case OP_REPEAT_INC_NG_SG:
 	    p += SIZE_MEMNUM;
 	    break;
 	  default:
@@ -2092,6 +2158,7 @@ match_at(regex_t* reg, UChar* str, UChar* end, UChar* sstart,
       GET_RELADDR_INC(addr, p);
       p += addr;
       STAT_OP_OUT;
+      CHECK_INTERRUPT_IN_MATCH_AT;
       continue;
       break;
 
@@ -2170,79 +2237,69 @@ match_at(regex_t* reg, UChar* str, UChar* end, UChar* sstart,
       break;
 
     case OP_REPEAT_INC:  STAT_OP_IN(OP_REPEAT_INC);
-      {
-	StackIndex si;
-
-	GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
-#ifdef USE_SUBEXP_CALL
-	if (reg->num_call > 0) {
-	  STACK_GET_REPEAT(mem, stkp);
-	  si = GET_STACK_INDEX(stkp);
-	}
-	else {
-	  si = repeat_stk[mem];
-	  stkp = STACK_AT(si);
-	}
-#else
-	si = repeat_stk[mem];
-	stkp = STACK_AT(si);
-#endif
-	stkp->u.repeat.count++;
-	if (stkp->u.repeat.count == reg->repeat_range[mem].upper) {
-	  /* end of repeat. Nothing to do. */
-	}
-	else if (stkp->u.repeat.count >= reg->repeat_range[mem].lower) {
-	  STACK_PUSH_ALT(p, s, sprev);
-	  p = stkp->u.repeat.pcode;
-	}
-	else {
-	  p = stkp->u.repeat.pcode;
-	}
-	STACK_PUSH_REPEAT_INC(si);
+      GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
+      si = repeat_stk[mem];
+      stkp = STACK_AT(si);
+  
+    repeat_inc:
+      stkp->u.repeat.count++;
+      if (stkp->u.repeat.count == reg->repeat_range[mem].upper) {
+        /* end of repeat. Nothing to do. */
       }
+      else if (stkp->u.repeat.count >= reg->repeat_range[mem].lower) {
+        STACK_PUSH_ALT(p, s, sprev);
+        p = STACK_AT(si)->u.repeat.pcode; /* Don't use stkp! */
+      }
+      else {
+        p = stkp->u.repeat.pcode;
+      }
+      STACK_PUSH_REPEAT_INC(si);
       STAT_OP_OUT;
+      CHECK_INTERRUPT_IN_MATCH_AT;
       continue;
       break;
 
+    case OP_REPEAT_INC_SG:  STAT_OP_IN(OP_REPEAT_INC_SG);
+      GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
+      STACK_GET_REPEAT(mem, stkp);
+      si = GET_STACK_INDEX(stkp);
+      goto repeat_inc;
+      break;
+
     case OP_REPEAT_INC_NG:  STAT_OP_IN(OP_REPEAT_INC_NG);
-      {
-	StackIndex si;
+      GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
+      si = repeat_stk[mem];
+      stkp = STACK_AT(si);
 
-	GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
-#ifdef USE_SUBEXP_CALL
-	if (reg->num_call > 0) {
-	  STACK_GET_REPEAT(mem, stkp);
-	  si = GET_STACK_INDEX(stkp);
-	}
-	else {
-	  si = repeat_stk[mem];
-	  stkp = STACK_AT(si);
-	}
-#else
-	si = repeat_stk[mem];
-	stkp = STACK_AT(si);
-#endif
-	stkp->u.repeat.count++;
-	if (stkp->u.repeat.count < reg->repeat_range[mem].upper) {
-	  if (stkp->u.repeat.count >= reg->repeat_range[mem].lower) {
-	    UChar* pcode = stkp->u.repeat.pcode;
+    repeat_inc_ng:
+      stkp->u.repeat.count++;
+      if (stkp->u.repeat.count < reg->repeat_range[mem].upper) {
+        if (stkp->u.repeat.count >= reg->repeat_range[mem].lower) {
+          UChar* pcode = stkp->u.repeat.pcode;
 
-	    STACK_PUSH_REPEAT_INC(si);
-	    STACK_PUSH_ALT(pcode, s, sprev);
-	  }
-	  else {
-	    p = stkp->u.repeat.pcode;
-	    STACK_PUSH_REPEAT_INC(si);
-	  }
-	}
-	else if (stkp->u.repeat.count == reg->repeat_range[mem].upper) {
-	  STACK_PUSH_REPEAT_INC(si);
-	}
+          STACK_PUSH_REPEAT_INC(si);
+          STACK_PUSH_ALT(pcode, s, sprev);
+        }
+        else {
+          p = stkp->u.repeat.pcode;
+          STACK_PUSH_REPEAT_INC(si);
+        }
+      }
+      else if (stkp->u.repeat.count == reg->repeat_range[mem].upper) {
+        STACK_PUSH_REPEAT_INC(si);
       }
       STAT_OP_OUT;
+      CHECK_INTERRUPT_IN_MATCH_AT;
       continue;
       break;
 
+    case OP_REPEAT_INC_NG_SG:  STAT_OP_IN(OP_REPEAT_INC_NG_SG);
+      GET_MEMNUM_INC(mem, p); /* mem: OP_REPEAT ID */
+      STACK_GET_REPEAT(mem, stkp);
+      si = GET_STACK_INDEX(stkp);
+      goto repeat_inc_ng;
+      break;
+
     case OP_PUSH_POS:  STAT_OP_IN(OP_PUSH_POS);
       STACK_PUSH_POS(s, sprev);
       STAT_OP_OUT;
@@ -2756,6 +2813,23 @@ onig_match(regex_t* reg, UChar* str, UChar* end, UChar* at, OnigRegion* region,
   UChar *prev;
   MatchArg msa;
 
+  if (ONIG_STATE(reg) == ONIG_STATE_NORMAL) {
+    reg->state++;  /* increment as search counter */
+    if (IS_NOT_NULL(reg->chain)) {
+      onig_chain_reduce(reg);
+      reg->state++;
+    }
+  }
+  else {
+    int n = 0;
+    while (ONIG_STATE(reg) < ONIG_STATE_NORMAL) {
+      if (++n > THREAD_PASS_LIMIT_COUNT)
+	return ONIGERR_OVER_THREAD_PASS_LIMIT_COUNT;
+      THREAD_PASS;
+    }
+    reg->state++;  /* increment as search counter */
+  }
+
   MATCH_ARG_INIT(msa, option, region, at);
 
   if (region
@@ -2772,6 +2846,8 @@ onig_match(regex_t* reg, UChar* str, UChar* end, UChar* at, OnigRegion* region,
     prev = onigenc_get_prev_char_head(reg->enc, str, at);
     r = match_at(reg, str, end, at, prev, &msa);
   }
+
+  reg->state--;  /* decrement as search counter */
   MATCH_ARG_FREE(msa);
   return r;
 }
